package cc.glsn.v15.ga;

import java.util.Random;

import java.nio.ByteBuffer;

import java.sql.ResultSet;

import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;
import java.io.ByteArrayOutputStream;
import java.util.TreeMap;

public class Entity
{
	private String id;
	private int op_count;
	private int instructions[];
	private int generation;
	private double fitness;
	private int last_mem[];


	public Entity(Random r, int ops, int max_opcode)
	{
		op_count = ops;
		
		instructions = new int[ops * 2];

		for(int i=0; i< ops; i++)
		{
			instructions[i*2] = r.nextInt(max_opcode);
			instructions[i*2+1] = r.nextInt();

		}

		generation=1;

		setID();
	
	}

	public Entity(int instructions[], int gen)
	{
		op_count = instructions.length / 2;
		this.instructions = instructions;
		generation = gen;
		setID();
	}

	public Entity(ResultSet r) throws java.sql.SQLException
	{
		id = r.getString("id");
		generation = r.getInt("generation");
		fitness = r.getDouble("fitness");
		byte[] blob = r.getBytes("instruction_data");
		op_count = blob.length / 4 / 2;

		ByteBuffer bb = ByteBuffer.wrap(blob);
		instructions = new int[op_count * 2];

		for(int i=0; i<blob.length/4; i++)
		{
			instructions[i] = bb.getInt();
		}

	}

	public int getGeneration()
	{
		return generation;
	}

	private int getIdx(int op)
	{
		if (op < 0) op = -op;

		return (op % op_count) * 2;
	}

	public void setOp(State s, int op)
	{
		int idx = getIdx(op);
		s.setOp(instructions[idx], instructions[idx+1]);
	}

	public static Entity mate(Random r, Entity e1, Entity e2)
	{
		int len = Math.min(e1.instructions.length, e2.instructions.length);
		int inst[] = new int[len];
		int i=0;

		TreeMap<Double, Integer[]> sections=
			new TreeMap<Double, Integer[]>();


		while(i < len)
		{
			int chunk_len = (int)Math.round(8.0 + r.nextGaussian()*3.0);
			int max = len - i;
			if (chunk_len < 1) chunk_len = 1;

			Entity src = e1;
			if (r.nextBoolean())
			{
				src = e2;
			}
			chunk_len = Math.min(chunk_len, len - i);

			Integer[] chunk=new Integer[chunk_len];

			for(int j=0; j<chunk_len; j++)
			{
				chunk[j] = src.instructions[i+j];
				inst[i+j] = src.instructions[i+j];
			}
			i += chunk_len;
			sections.put(i * (r.nextGaussian() + 0.5), chunk);


		}

		int reorder[] = new int[len];
		i=0;
		for(Integer[] I : sections.values())
		{
			for(int j=0; j<I.length; j++)
			{
				reorder[i] = I[j];
				i++;
			}

		}



		return new Entity(inst, Math.max(e1.generation, e2.generation) + 1);
		//return new Entity(reorder, Math.max(e1.generation, e2.generation) + 1);

	}
	

	public byte[] getInstructionByteArray()
	{
		byte[] buff= new byte[instructions.length * 4];
		ByteBuffer bb = ByteBuffer.wrap(buff);
		for(int i=0; i<instructions.length; i++)
		{
			bb.putInt(instructions[i]);
		}
		return buff;

	}

	private void setID()
	{
		id = cc.glsn.Util.SHA1(getInstructionByteArray()).substring(0,12);
	}

	public String getID()
	{
		return id;
	}

	public void setFitness(double fit)
	{
		fitness = fit;
	}
	public double getFitness()
	{
		return fitness;
	}

	public void setLastMem(int mem[])
	{
		last_mem = mem;

	}
	public int[] getLastMem()
	{
		return last_mem;
	}

	public byte[] getImageBytes()
	{
		if (last_mem == null) return null;

		BufferedImage bi = new BufferedImage(256,256,BufferedImage.TYPE_INT_ARGB);

		for(int i=0; i<256; i++)
		for(int j=0; j<256; j++)
		{
			bi.setRGB(i,j,last_mem[i*256+j]);
		}

		ByteArrayOutputStream bout = new ByteArrayOutputStream();

		try
		{
			ImageIO.write(bi, "PNG", bout);
		}
		catch(java.io.IOException e)
		{
			e.printStackTrace();
			return null;
		}

		return bout.toByteArray();
	}

}
