package cc.glsn.v15.ga;

import java.util.Map;
import java.util.ArrayList;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Random;
import java.text.DecimalFormat;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.PreparedStatement;
import java.sql.DriverManager;



public class Simulate
{

	//Options
	String project_name;
	int instruction_size;
	int mem_size;
	long max_ops_per_run;
	int target_pop;
	int fresh_per_pass;
	int db_pass;
	int offspring;
	FitnessTest ft;


	//Run time kelp
	Random rnd;
	Map<Integer, Op> inst_map;
	TreeMap<Double, Entity> pool;
	TreeSet<String> ids_in_db;
	Connection conn;

	public Simulate(String project_name, int instruction_size, int mem_size, long max_ops_per_run,int target_pop, int fresh_per_pass, int db_pass, int offspring, FitnessTest ft)
	{
		this.project_name = project_name;
		this.instruction_size = instruction_size;
		this.mem_size = mem_size;
		this.max_ops_per_run = max_ops_per_run;
		this.target_pop = target_pop;
		this.fresh_per_pass = fresh_per_pass;
		this.db_pass = db_pass;
		this.offspring = offspring;
		this.ft = ft;

		rnd = new Random();
		inst_map = loadOpMap();
		pool = new TreeMap<Double, Entity>();
		ids_in_db = new TreeSet<String>();


	}

	public void goHuman()
	{
		int pass=0;
		while(true)
		{
			try
			{
				if (conn == null)
				{
					openDB();
				}

				double total_votes=getDBVotes();
				double total_entries=getDBEntries();
				long boundry = (long)Math.round((total_votes / total_entries) / 2.0);

				boundry = Math.max(7, boundry);

				System.out.println("Total votes: " + total_votes + " Entries: " + total_entries);
				System.out.println("Each needs: " + boundry);


				long ready=getDBEntriesAbove(boundry);
				long not_ready=getDBEntriesBelow(boundry);

				System.out.println("Ready/Not: " + ready + "/" + not_ready);
				



				if (not_ready < target_pop)
				{
					long to_add = target_pop - not_ready;

					System.out.println("Making new: " + to_add);
					pool.clear();

					//loadFromDB(0.0,target_pop*3);
					loadFromDBByVotes(target_pop);

					ArrayList<Entity> al = new ArrayList<Entity>();
					al.addAll(pool.values());

					for(int i=0; i<fresh_per_pass; i++)
					{
						Entity e = new Entity(rnd, instruction_size, inst_map.size());
						al.add(e);
					}
				
					to_add*=2;

					for(long x=0; x<to_add; x++)
					{
						int i = rnd.nextInt(al.size());

						int j = i;
						while (i == j)
						{
							j = rnd.nextInt(al.size());
						}
						Entity e = Entity.mate(rnd, al.get(i), al.get(j));
						double f = findFitness(e);

						//Only save new ones if they don't completely suck
						if (f > 1000.0 )
						{
							pool.put(f, e);
						}
					}
					saveToDB(0.0);
				

				}
				if (ready > target_pop)
				{
					pruneDBAbove(boundry);

				}




			}
			catch(java.sql.SQLException e)
			{
				e.printStackTrace();
				try{if (conn != null) conn.close();}catch(Exception e2){}
				conn = null;
			}

			try{Thread.sleep(120000);}catch(Exception e){}

		}

	}

	public void goInjectFresh()
	{
		int pass=0;

		while(true)
		{
			System.out.println("Starting Pass " + pass);
			//Remove unfit
			while(pool.size() > target_pop)
			{
				pool.remove(pool.firstKey());
			}

			//Maybe do DB

			if (pass % db_pass == 0)
			{
				try
				{
					System.out.println("  Doing db operations");
					if (conn == null)
					{
						openDB();
					}
					saveToDB(0.0);
					//pruneDB();

					pool.clear();
					if (pass >0)
					{
						return;
					}


					
					
				}
				catch(java.sql.SQLException e)
				{
					e.printStackTrace();
					try{if (conn != null) conn.close();}catch(Exception e2){}
					conn = null;
				}

			}
			
			//Remove unfit
			while(pool.size() > target_pop)
			{
				pool.remove(pool.firstKey());
			}

			addFreshBlood();


			ArrayList<Entity> al = new ArrayList<Entity>();
			al.addAll(pool.values());

			for(int o=0; o<offspring; o++)
			{
				//Mate everything to at least one thing
				for(int i=0; i<al.size(); i++)
				{
					int j = i;
					while (i == j)
					{
						j = rnd.nextInt(al.size());
					}
					Entity e = Entity.mate(rnd, al.get(i), al.get(j));
					pool.put(findFitness(e), e);

				}
			}
			
			pass++;



		}


	}

	public void goComputer()
	{

		int pass=0;
		while(true)
		{
			System.out.println("Starting Pass " + pass);
			//Remove unfit
			while(pool.size() > target_pop)
			{
				pool.remove(pool.firstKey());
			}

			//Maybe do DB

			if (pass % db_pass == 0)
			{
				try
				{
					System.out.println("  Doing db operations");
					if (conn == null)
					{
						openDB();
					}
					/*double min=0.0;
					if (pool.size() > 0)
					{
						min = pool.lastKey()-1.0;
					}
					loadFromDB(min, target_pop);*/
					//Remove unfit
					while(pool.size() > target_pop)
					{
						pool.remove(pool.firstKey());
					}
					saveToDB(0.0);
					//pruneDB();

					
					
				}
				catch(java.sql.SQLException e)
				{
					e.printStackTrace();
					try{if (conn != null) conn.close();}catch(Exception e2){}
					conn = null;
				}

			}
			
			//Remove unfit
			while(pool.size() > target_pop)
			{
				pool.remove(pool.firstKey());
			}

			addFreshBlood();


			ArrayList<Entity> al = new ArrayList<Entity>();
			al.addAll(pool.values());

			for(int o=0; o<offspring; o++)
			{
				//Mate everything to at least one thing
				for(int i=0; i<al.size(); i++)
				{
					int j = i;
					while (i == j)
					{
						j = rnd.nextInt(al.size());
					}
					Entity e = Entity.mate(rnd, al.get(i), al.get(j));
					pool.put(findFitness(e), e);

				}
			}
			
			pass++;



		}

	}


	private void addFreshBlood()
	{
			//Add fresh blood
			while(pool.size() < target_pop + fresh_per_pass)
			{
				Entity e = new Entity(rnd, instruction_size, inst_map.size());
				pool.put(findFitness(e), e);
			}

	}
	public double findFitness(Entity e)
	{
		int[] mem = new int[mem_size];
		long op = run(e,mem);


		int[] mem_copy = new int[mem_size];
		for(int i=0; i<mem_size; i++)
		{
			mem_copy[i] = mem[i];
		}

		double score = ft.findFitness(mem);

		e.setFitness(score);

		//System.out.println("  id: " + e.getID() + " gen: " + getNumber(e.getGeneration(), 4) + " ops: " + getNumber(op,9) + " score: " + getNumber(score,12) );

		return (rnd.nextDouble() + score);
	}

	private static String getNumber(double n, int len)
	{
		DecimalFormat df = new DecimalFormat("0.0");
		String s = df.format(n);
		while(s.length() < len) s=" " + s;
		return s;

	}
	private static String getNumber(long n, int len)
	{
		String s = "" + n;
		while(s.length() < len) s=" " + s;
		return s;
	}

	public int[] singleRun(Entity e)
	{
		int[] mem = new int[mem_size];

		run(e, mem);

		return mem;

	}

	public long run(Entity e, int[] mem)
	{
		State s = new State(mem);

		long ops=0;
		while((!s.isTerminated()) && (ops < max_ops_per_run))
		{
			s.loadInst(e);

			s.runOp(inst_map);

			ops++;
		}

		e.setLastMem(mem);

		return ops;
		
	}

	public static Map<Integer, Op> loadOpMap()
	{
		TreeMap<Integer, Op> im = new TreeMap<Integer, Op>();

		im.put(0, new OpNull());
		im.put(1, new OpJumpIfZero());
		im.put(2, new OpLoadLiteralR1());
		im.put(3, new OpLoadLiteralR2());
		im.put(4, new OpStore());
		im.put(5, new OpLoad());
		im.put(6, new OpSwap());

		im.put(7, new OpIInc());
		im.put(8, new OpIAdd());
		im.put(9, new OpINeg());
		im.put(10, new OpIMult());
		im.put(11, new OpIDiv());
		
		im.put(12, new OpFAdd());
		im.put(13, new OpFNeg());
		im.put(14, new OpFMult());
		im.put(15, new OpFDiv());

		im.put(16, new OpI2F());
		im.put(17, new OpF2I());
		im.put(18, new OpIComp());
		im.put(19, new OpFComp());

		im.put(20, new OpTerminate());
		im.put(21, new OpFSin());
		im.put(22, new OpFCos());
		im.put(23, new OpFLog());
		im.put(24, new OpFExp());

		return im;
	}

	private void openDB() throws java.sql.SQLException
	{
		conn =  DB.open();

	}

	private long getDBVotes() throws java.sql.SQLException
	{
		PreparedStatement ps = conn.prepareStatement("select sum(votes) from entity where project=?");
		ps.setString(1, project_name);
		ResultSet r = ps.executeQuery();

		long n=-1;
		if (r.next())
		{
			n = r.getLong("sum(votes)");
		}

		r.close();
		ps.close();
		
		return n;
	}
	
	private long getDBEntries() throws java.sql.SQLException
	{
		PreparedStatement ps = conn.prepareStatement("select count(id) from entity where project=?");
		ps.setString(1, project_name);
		ResultSet r = ps.executeQuery();

		long n=-1;
		if (r.next())
		{
			n = r.getLong("count(id)");
		}

		r.close();
		ps.close();

		return n;
	}
	private long getDBEntriesAbove(long votes) throws java.sql.SQLException
	{
		PreparedStatement ps = conn.prepareStatement("select count(id) from entity where project=? and votes>=?");
		ps.setString(1, project_name);
		ps.setLong(2, votes);
		ResultSet r = ps.executeQuery();

		long n=-1;
		if (r.next())
		{
			n = r.getLong("count(id)");
		}

		r.close();
		ps.close();

		return n;
	}
	
	private long getDBEntriesBelow(long votes) throws java.sql.SQLException
	{
		PreparedStatement ps = conn.prepareStatement("select count(id) from entity where project=? and votes<?");
		ps.setString(1, project_name);
		ps.setLong(2, votes);
		ResultSet r = ps.executeQuery();

		long n=-1;
		if (r.next())
		{
			n = r.getLong("count(id)");
		}

		r.close();
		ps.close();

		return n;

	}




	private void loadFromDB(double minfitness, int limit)
		 throws java.sql.SQLException
	{

		TreeSet<String> loaded=new TreeSet<String>();
		for(Entity e : pool.values())
		{
			loaded.add(e.getID());
		}	

		//ids_in_db.clear();

		int n=0;

		PreparedStatement ps = conn.prepareStatement("select id,instruction_data,fitness,generation from entity where project=? and fitness>=? order by fitness desc limit ?");

		ps.setString(1, project_name);
		ps.setDouble(2, minfitness);
		ps.setInt(3, limit);

		ResultSet r = ps.executeQuery();

		while(r.next())
		{
			Entity e = new Entity(r);
			double fitness = e.getFitness();	

			if (!loaded.contains(e.getID()))
			{
				pool.put(fitness + rnd.nextDouble(), e);
				n++;
			}

			ids_in_db.add(e.getID());
			
		}


		r.close();
		ps.close();

		System.out.println("    " + n + " loaded from db");

	}
	private void loadFromDBByVotes(int limit)
		throws java.sql.SQLException
	{
		TreeSet<String> loaded=new TreeSet<String>();
		for(Entity e : pool.values())
		{
			loaded.add(e.getID());
		}	

		//ids_in_db.clear();

		int n=0;

		PreparedStatement ps = conn.prepareStatement("select id,instruction_data,fitness,generation from entity where project=?  order by votes_up/votes desc limit ?");

		ps.setString(1, project_name);
		ps.setInt(2, limit);

		ResultSet r = ps.executeQuery();

		while(r.next())
		{
			Entity e = new Entity(r);
			double fitness = e.getFitness();	

			if (!loaded.contains(e.getID()))
			{
				pool.put(fitness + rnd.nextDouble(), e);
				n++;
			}

			ids_in_db.add(e.getID());
			
		}


		r.close();
		ps.close();

		System.out.println("    " + n + " loaded from db");

	}
	private void saveToDB(double minfitness)
		 throws java.sql.SQLException
	{
		int n=0;
		for(Entity e : pool.values())
		{
			String id = e.getID();
			if ((!ids_in_db.contains(id)) && (e.getFitness() >= minfitness))
			{
				PreparedStatement ps = conn.prepareStatement
					("insert into entity (id, project, instruction_data, fitness, generation, image) values (?,?,?,?,?,?)");
				ps.setString(1, id);
				ps.setString(2, project_name);
				ps.setBytes(3, e.getInstructionByteArray());
				ps.setDouble(4, e.getFitness());
				ps.setInt(5, e.getGeneration());

				byte[] b=e.getImageBytes();
				//System.out.println("id: " + id +" size: " + b.length);

				ps.setBytes(6, b);

				ps.execute();
				ps.close();

				n++;
				
				ids_in_db.add(id);
			}
		}
		
		System.out.println("    " + n + " saved to db");

	}

	private double getMinDBFitness()
		throws java.sql.SQLException
	{
		double lowest_seen = 0.00;
		int seen=0;
		{
			PreparedStatement ps = conn.prepareStatement("select fitness from entity where project=? order by fitness desc");

			ps.setString(1, project_name);

			ResultSet r= ps.executeQuery();

			while(r.next() && (seen < target_pop))
			{
				lowest_seen = r.getDouble("fitness");
				seen++;
			}

			r.close();

			ps.close();
		}
		return lowest_seen;


	}

	private void pruneDB()
		 throws java.sql.SQLException
	{
		double lowest_seen = getMinDBFitness();
		{

			PreparedStatement ps = conn.prepareStatement("delete from entity where project=? and fitness < ?");
			ps.setString(1, project_name);
			ps.setDouble(2, lowest_seen);
			int n = ps.executeUpdate();
			ps.close();

			System.out.println("    " + n + " pruned from db");
		}

	}
	
	private void pruneDBAbove(long votes)
		 throws java.sql.SQLException
	{

		double lowest_seen = 0.00;
		int seen=0;
		{
			PreparedStatement ps = conn.prepareStatement("select votes_up/votes from entity where project=? and votes>=? order by votes_up/votes desc");

			ps.setString(1, project_name);
			ps.setLong(2, votes);

			ResultSet r= ps.executeQuery();

			while(r.next() && (seen < target_pop))
			{
				lowest_seen = r.getDouble("votes_up/votes");
				seen++;
			}

			r.close();

			ps.close();
		}

		if (seen >= target_pop)
		{
			lowest_seen -= 0.01;

			System.out.println("delete from entity where project=? and votes>="+votes+" and votes_up/votes < " + lowest_seen);
			
			PreparedStatement ps = conn.prepareStatement("delete from entity where project=? and votes>=? and votes_up/votes < ?");
			ps.setString(1, project_name);
			ps.setLong(2, votes);
			ps.setDouble(3, lowest_seen);
			int n = ps.executeUpdate();
			ps.close();

			System.out.println("    " + n + " pruned from db");
		}

	}





}


